import enum
import logging
import os
from pathlib import Path
from typing import Optional

import typer
from rich.logging import RichHandler
from rich.pretty import pprint
from typer.core import TyperGroup

import mmuxer

from .cli.folder import app as folder_app
from .cli.run import monitor as monitor_cmd
from .cli.run import tidy as tidy_cmd
from .cli.sieve_export import sieve_export
from .config_state import state
from .utils import config_file_typer_option


class OrderCommands(TyperGroup):
    def list_commands(self, ctx: typer.Context):
        """Return list of commands in the order they appear."""
        return list(self.commands)  # get commands using self.commands


class LogLevel(enum.Enum):
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


def main_callback(
    log_level: LogLevel = typer.Option("info", case_sensitive=False),
    journald_logger: Optional[str] = typer.Option(
        None, help="If provided, log to journald instead of stdout (with given logger name)"
    ),
):
    if journald_logger:
        from cysystemd import journal

        handler = journal.JournaldLogHandler(identifier=journald_logger)
    elif os.isatty(0):
        handler = RichHandler(
            rich_tracebacks=True, show_time=False, show_path=(log_level == LogLevel.DEBUG)
        )
    else:
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(levelname)s - %(message)s")
        handler.setFormatter(formatter)

    logging.basicConfig(
        level=log_level.name,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[handler],
    )


app = typer.Typer(
    no_args_is_help=True,
    cls=OrderCommands,
    callback=main_callback,
    help=f"Mail Muxer - Version {mmuxer.__version__}",
)


app.command(rich_help_panel="Main commands")(monitor_cmd)
app.command(rich_help_panel="Main commands")(tidy_cmd)
app.command(rich_help_panel="Main commands")(sieve_export)


@app.command(rich_help_panel="Util commands")
def check(config_file: Path = config_file_typer_option):
    """Load the config and connect to the IMAP server."""
    state.load_config_file(config_file)
    state.create_mailbox()
    pprint(state.rules)


app.add_typer(folder_app, rich_help_panel="Util commands")

if __name__ == "__main__":
    app()



from typing import List, Literal, Union

import boolean

from mmuxer.models.common import BaseModel
from mmuxer.models.condition import All, Any, BaseCondition, Condition, Not, is_base_condition

algebra = boolean.BooleanAlgebra()
Clause = Union[algebra.Symbol, algebra.NOT, algebra.OR, algebra.AND]


def remove_singleton_conditions(condition: Condition) -> Condition:
    if is_base_condition(condition):
        return condition
    elif isinstance(condition, Not):
        return Not(NOT=remove_singleton_conditions(condition.NOT))
    elif isinstance(condition, Any):
        if len(condition.ANY) == 1:
            return remove_singleton_conditions(condition.ANY[0])
        else:
            return Any(ANY=[remove_singleton_conditions(cond) for cond in condition.ANY])
    elif isinstance(condition, All):
        if len(condition.ALL) == 1:
            return remove_singleton_conditions(condition.ALL[0])
        else:
            return All(ALL=[remove_singleton_conditions(cond) for cond in condition.ALL])


def parse_condition(condition: Condition) -> Clause:
    if is_base_condition(condition):
        return algebra.Symbol(condition)
    elif isinstance(condition, Not):
        return algebra.NOT(parse_condition(condition.NOT))
    elif isinstance(condition, Any):
        return algebra.OR(*(parse_condition(cond) for cond in condition.ANY))
    elif isinstance(condition, All):
        return algebra.AND(*(parse_condition(cond) for cond in condition.ALL))


def to_condition(expression: Clause) -> Condition:
    if isinstance(expression, algebra.Symbol):
        return expression.obj
    if isinstance(expression, algebra.NOT):
        return Not(NOT=to_condition(expression.args[0]))
    elif isinstance(expression, algebra.AND):
        return All(ALL=[to_condition(e) for e in expression.args])
    elif isinstance(expression, algebra.OR):
        return Any(ANY=[to_condition(e) for e in expression.args])
    else:
        raise Exception(f"Unexpected expression {expression}")


def to_dnf(condition: Condition):
    normalized_condition = remove_singleton_conditions(condition)
    clause = parse_condition(normalized_condition)
    dnf_clause = algebra.normalize(clause, algebra.OR)
    return to_condition(dnf_clause)


#
def depth(condition: Condition) -> int:
    """We expect the condition to be in normal form, with Not only
    applied to BaseConditions"""
    if is_base_condition(condition) or isinstance(condition, Not):
        return 0
    elif isinstance(condition, Any):
        return 1 + max(depth(cond) for cond in condition.ANY)
    elif isinstance(condition, All):
        return 1 + max(depth(cond) for cond in condition.ALL)


class SieveCondition(BaseModel):
    type: Union[Literal["allof"], Literal["anyof"]]
    conditions: List[Union[BaseCondition, Not]]

    def dump(self):
        conditions_to_sieve = ", ".join(condition.to_sieve() for condition in self.conditions)
        return f"if {self.type} ({conditions_to_sieve})"


def to_sieve_conditions_flat(condition: Condition) -> List[SieveCondition]:
    condition_dnf = to_dnf(condition)
    if is_base_condition(condition_dnf) or isinstance(condition_dnf, Not):
        return [SieveCondition(type="anyof", conditions=[condition_dnf])]
    elif isinstance(condition_dnf, All):
        return [SieveCondition(type="allof", conditions=condition_dnf.ALL)]
    elif isinstance(condition_dnf, Any):
        first_level_conds = [cond for cond in condition_dnf.ANY if depth(cond) == 0]
        second_level_conds = [cond for cond in condition_dnf.ANY if depth(cond) == 1]
        # When in dnf, conditions of level 1 are All conditions
        assert all(isinstance(all_conds, All) for all_conds in second_level_conds)

        res = []
        if first_level_conds:
            res.append(SieveCondition(type="anyof", conditions=first_level_conds))
        if second_level_conds:
            res.extend(
                SieveCondition(type="allof", conditions=all_conds.ALL)
                for i, all_conds in enumerate(second_level_conds)
            )
        return res
    else:
        raise Exception(f"Unexpected condition {condition_dnf}")


def to_sieve_condition_rec(condition: Condition, indent_n: int) -> str:
    indent = "  " * indent_n
    if isinstance(condition, Not):
        inner = to_sieve_condition_rec(condition.NOT, indent_n)
        return f"not {inner}"
    if isinstance(condition, All):
        inner_conds = (to_sieve_condition_rec(cond, indent_n + 1) for cond in condition.ALL)
        inner = f",\n".join(inner_conds)
        return f"{indent}allof (\n{inner}\n{indent})"
    if isinstance(condition, Any):
        inner_conds = (to_sieve_condition_rec(cond, indent_n + 1) for cond in condition.ANY)
        inner = f",\n".join(inner_conds)
        return f"{indent}anyof (\n{inner}\n{indent})"

    return f"{indent}{condition.to_sieve()}"


def to_sieve_conditions(condition: Condition) -> str:
    return "if " + to_sieve_condition_rec(condition, 0)



from enum import Enum


class ComparisonOperator(Enum):
    CONTAINS = "CONTAINS"
    EQUALS = "EQUALS"

    @property
    def sieve(self):
        if self == ComparisonOperator.CONTAINS:
            return ":contains"
        if self == ComparisonOperator.EQUALS:
            return ":is"
        raise Exception(f"Unhandled operator {self}")

    def eval(self, operand: str, value: str) -> bool:
        if self == ComparisonOperator.CONTAINS:
            return operand.lower() in value.lower()
        if self == ComparisonOperator.EQUALS:
            return operand.lower() == value.lower()
        raise Exception(f"Unhandled operator {self}")


class Flag(Enum):
    SEEN = "SEEN"
    ANSWERED = "ANSWERED"
    FLAGGED = "FLAGGED"
    DELETED = "DELETED"
    DRAFT = "DRAFT"

    @property
    def imap(self):
        return "\\" + self.name.capitalize()

    @property
    def sieve(self):
        return "\\\\" + self.name.capitalize()



import logging
from abc import abstractmethod
from collections.abc import Sequence
from typing import ForwardRef, List, Union

from imap_tools import MailMessage

from .common import BaseModel
from .enums import ComparisonOperator


class IBaseCondition(BaseModel):
    class Config:
        frozen = True

    operator: ComparisonOperator = ComparisonOperator.CONTAINS

    @abstractmethod
    def get_value(self, message: MailMessage) -> str:
        """Extract value from the message, used as value to which the rule is applied to."""
        pass

    @abstractmethod
    def get_operand(self) -> Union[str, Sequence[str]]:
        """Used to have a common function returning the operand of the rule across subclasses."""
        pass

    def eval(self, message: MailMessage) -> bool:
        """Evaluate the rule"""
        value = self.get_value(message)
        operand = self.get_operand()
        logging.debug("Eval %s <%s> %s", operand, self.operator, value)
        if isinstance(operand, str):
            operands = [operand]
        else:
            operands = operand

        return any(self.operator.eval(op, value) for op in operands)

    def to_sieve(self) -> str:
        """Used to render sieve files
        We include here the one used for header conditions
        """
        operand = self.get_operand()
        if isinstance(operand, str):
            operand_str = f'"{operand}"'
        else:
            operand_str = "[" + ", ".join(f'"{op}"' for op in operand) + "]"

        return f'header {self.operator.sieve} "{self.__class__.__name__.lower()}" {operand_str}'

    def __lt__(self, other):
        """Comparison: used for using with boolean.py"""
        return repr(self).__lt__(repr(other))


class From(IBaseCondition):
    FROM: Union[str, frozenset[str]]

    def get_value(self, message: MailMessage):
        return message.from_

    def get_operand(self) -> Union[str, Sequence[str]]:
        return self.FROM

    def __rich_repr__(self):
        yield self.operator.name, self.FROM


class To(IBaseCondition):
    TO: Union[str, frozenset[str]]

    def get_value(self, message: MailMessage):
        return " ".join(message.to)

    def get_operand(self) -> Union[str, Sequence[str]]:
        return self.TO

    def __rich_repr__(self):
        yield self.operator.name, self.TO


class Subject(IBaseCondition):
    SUBJECT: Union[str, frozenset[str]]

    def get_value(self, message: MailMessage):
        return message.subject

    def get_operand(self) -> Union[str, Sequence[str]]:
        return self.SUBJECT

    def __rich_repr__(self):
        yield self.operator.name, self.SUBJECT


class Body(IBaseCondition):
    BODY: str

    def get_value(self, message: MailMessage):
        return message.text + message.html

    def get_operand(self) -> str:
        return self.BODY

    def __rich_repr__(self):
        yield self.operator.name, self.BODY

    def to_sieve(self) -> str:
        """Used to render sieve files"""
        return f'body :text {self.operator.sieve} "{self.get_operand()}'


BaseCondition = Union[From, To, Subject, Body]


def is_base_condition(obj):
    """Useful for isinstance tests with python < 3.10"""
    return isinstance(obj, From) or isinstance(obj, To) or isinstance(obj, Subject)


All = ForwardRef("All")  # type: ignore
Any = ForwardRef("Any")  # type: ignore
Not = ForwardRef("Not")  # type: ignore


class All(BaseModel):
    ALL: List[Union[BaseCondition, All, Any, Not]]

    def eval(self, message: MailMessage):
        return all(operand.eval(message) for operand in self.ALL)

    def __rich_repr__(self):
        for item in self.ALL:
            yield item


class Any(BaseModel):
    ANY: List[Union[BaseCondition, All, Any, Not]]

    def eval(self, message: MailMessage):
        return any(operand.eval(message) for operand in self.ANY)

    def __rich_repr__(self):
        for item in self.ANY:
            yield item


class Not(BaseModel):
    NOT: Union[BaseCondition, All, Any, Not]

    def eval(self, message: MailMessage):
        return not self.NOT.eval(message)

    def to_sieve(self) -> str:
        """Used to render sieve files"""
        return f"not {self.NOT.to_sieve()}"


All.update_forward_refs()
Any.update_forward_refs()
Not.update_forward_refs()

Condition = Union[BaseCondition, All, Any, Not]



from typing import Optional

from pydantic_settings import BaseSettings, SettingsConfigDict

from mmuxer.models.common import BaseModel


def in_container():
    """Returns: True iff running in a container"""
    with open("/proc/1/cgroup") as ifh:
        value = ifh.read()
        return "docker" in value or "kubepod" in value


class Settings(BaseModel, BaseSettings):
    server: str
    username: str
    password: str
    ssl_ciphers: Optional[str] = None
    imap_wait_timeout: int = 60
    sieve_folder_prefix: str = ""  # folder prefix used when generating sieve rules
    sieve_folder_separator: str = "/"  # folder separator used when generating sieve rules
    fetch_batch_size: int = 1000  # batch size used when fetching messages in bulk

    model_config = SettingsConfigDict(
        env_file=".env", secrets_dir="/run/secrets" if in_container() else None
    )
